/**
 * file             : ccd.c
 * author           : Ingo A. Kubbilun <ingo@pyrillion.org>
 * license          : GNU GENERAL PUBLIC LICENSE 2 (or later)
 * 
 * description      : This is the C compiler delegate ccd, which
 *                    makes the compilation of the Linux kernel
 *                    using Intel(R) C/C++ Compiler for Linux
 *                    possible.
 *
 *                    See ccd(1) for details.
 *
 * revision history :
 * ------------------
 * 
 * version  date        comments
 * ---------------------------------------------------------------------
 * 0.1      04/01/2004  completely rewritten based on gccicc (the former
 *                      compiler delegate) - Ingo A. Kubbilun
**/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>

extern int ICCAfterBurner ( char *argv0, char **argv );
typedef int (*PCHILDPROC)(char *, char **);
int DoExecute ( char *argv0, char **argv, PCHILDPROC pChildProc );

       char                 szCompiler[64];               // compiler to use, e.g. gcc or icc
       int                  bIsICC            = 0;        // is it the Intel Compiler for Linux (icc) ?
       int                  bAssem            = 0;
       int                  bPreproc          = 0;
       int                  bQuiet            = 0;
       int                  bCompileOnly      = 0;
       int                  bOutputGiven      = 0;
       int                  bInputGiven       = 0;
       int                  bInputIsAssembler = 0;
       int                  bNoCommon         = 0;
       int                  bRemoveForceJunk  = 0;
       int                  bOptInline        = 0;
       int                  bSafeSSE          = 0;
       int                  iProfGenPos       = 0;
       int                  bDetectasmBug1    = 0;
static int                  bNo2ndTry         = 0;
       int                  iCSwitchPos;
       int                  iOutputPos;
       int                  iInputPos;
       char                 szOutputFile[256];
       char                 szInputFile[256];
       char                 szAssemblerFile[256];

static void getenvars ( void )
{
  char                     *p;

  p = getenv("CCC");

  if (p)
    strcpy(szCompiler,p);
  else
  {
    p = getenv("CC");
    if (p)
      strcpy(szCompiler,p);
    else
      strcpy(szCompiler,"gcc");
  }

  if (!strcmp(szCompiler,"icc"))
    bIsICC = 1;
}

int DoExecute ( char *argv0, char **argv, PCHILDPROC pChildProc )
{
  pid_t         pid,wait_pid;
  int           status=1;

  pid=fork();
  if (!pid) // child
  {
    return pChildProc(argv0,argv);
  }
  else
  {
    while ((wait_pid=wait4(pid,&status,0,NULL))<0)
    {
      if (errno!=EINTR) break;
    }
    if (wait_pid!=pid)
    {
      fprintf(stderr,"CCC FATAL ERROR: cannot execute child process. ABORTING.\n");
      fflush(stderr);
      exit(1);
    }
    if (WIFSIGNALED(status))
    {
      fprintf(stderr,"CCC FATAL ERROR: unhandled signal in child process. ABORTING.\n");
      fflush(stderr);
      exit(1);
    }
  }
  return WEXITSTATUS(status);
}
 
static int PreprocessFile ( char *argv0, char **argv )
{
  char          szNewOutput[256];

  sprintf(szNewOutput,"%s.i",szInputFile);

  argv[iOutputPos]     = szNewOutput;
  argv[iCSwitchPos][1] = 'E';

  execvp(argv0,argv);
  return 1;
}

static int AssembleFile ( char *argv0, char **argv )
{
  char          szNewOutput[256];

  sprintf(szNewOutput,"%s.S",szInputFile);

  argv[iOutputPos]     = szNewOutput;
  argv[iCSwitchPos][1] = 'S';

  execvp(argv0,argv);
  return 1;
}

static int CompileFile ( char *argv0, char **argv )
{
  if ((bIsICC) && ((bNoCommon) || (bRemoveForceJunk) || (bOptInline) || (bSafeSSE)))
    exit(ICCAfterBurner(argv0,argv));
  else
    execvp(argv0,argv);

  return 1;
}

int main ( int argc, char *argv[] )
{
  int                     i=1,j=1,iRetVal;
  int                     newargc = argc;
  char                   *px;

  getenvars();

#ifdef _DEBUG
  fprintf(stdout,"\nCALL: ");
  for (int k=0;k<argc;k++) fprintf(stdout,"%s ",argv[k]);
  fprintf(stdout,"\n");
#endif

  while (i<argc)
  {
    if (!strcmp(argv[i],"--assem"))
    {
      bAssem=1;
      newargc--;
      i++;
    }
    else
    if (!strcmp(argv[i],"--preproc"))
    {
      bPreproc=1;
      newargc--;
      i++;
    }
    else
    if (!strcmp(argv[i],"--quiet"))
    {
      bQuiet=1;
      newargc--;
      i++;
    }
    else
    if (!strcmp(argv[i],"--no-common"))
    {
      bNoCommon=1;
      newargc--;
      i++;
    }
    else
    if (!strcmp(argv[i],"--remove-force-junk"))
    {
      bRemoveForceJunk=1;
      newargc--;
      i++;
    }
    else
    if (!strcmp(argv[i],"--no-2ndtry"))
    {
      bNo2ndTry=1;
      newargc--;
      i++;
    }
    else
    if (!strcmp(argv[i],"--opt-inline"))
    {
      bOptInline=1;
      newargc--;
      i++;
    }
    else
    if (!strcmp(argv[i],"--safe-sse"))
    {
      bSafeSSE=1;
      newargc--;
      i++;
    }
    else
    if (!strcmp(argv[i],"--detect-asmbug1"))
    {
      bDetectasmBug1=1;
      newargc--;
      i++;
    }
    else
    {
      if (!memcmp(argv[i],"-prof_gen",9))
      {
        iProfGenPos = j;
      }
      else
      if (!strcmp(argv[i],"-c"))
      {
        iCSwitchPos  = j;
        bCompileOnly = 1;
      }
      else
      if ((!strcmp(argv[i],"-o")) && (i<(argc-1)) && (strstr(argv[i+1],".o")))
      {
        bOutputGiven=1;
        strcpy(szOutputFile,argv[i+1]);
        argv[j++]  = argv[i++];
        iOutputPos = j;
      }
      else
      if (argv[i][0]!='-')
      {
        if (bInputGiven)
          bInputGiven++;
        else
        {
          if (strstr(argv[i],".c"))
          {
            bInputGiven = 1;
            strcpy(szInputFile,argv[i]);
            iInputPos = j;
          }
          else
          if (strstr(argv[i],".S"))
          {
            bInputGiven       = 1;
            bInputIsAssembler = 1;
            strcpy(szInputFile,argv[i]);
            iInputPos = j;
          }
        }
      }
      argv[j++] = argv[i++];
    }
  }
  argv[j]=NULL;
  strcpy(argv[0],szCompiler);
/*
  // disable prof_gen for boot stuff and intlpgo itself

  if ((iProfGenPos) && (bInputGiven==1))
  {
    if ((strstr(szInputFile,"arch/i386/boot")) || (strstr(szInputFile,"arch/i386/intlpgo")))
    {
      if (iOutputPos>iProfGenPos) iOutputPos--;
      if (iCSwitchPos>iProfGenPos) iCSwitchPos--;
      if (iInputPos>iProfGenPos) iInputPos--;
      while (argv[iProfGenPos])
      {
        argv[iProfGenPos] = argv[iProfGenPos+1];
        iProfGenPos++;
      }
    }
  }
*/
  if (bInputIsAssembler)  // assembler files are always processed by gcc
  {
    strcpy(argv[0],"gcc");
  }
  else
  {
    if ((bInputGiven==1) && (bOutputGiven) && (bCompileOnly))
    {
      if ((bIsICC) && ((bNoCommon) || (bRemoveForceJunk) || (bOptInline) || (bSafeSSE)))
      {
        px = strrchr(szInputFile,'/');
        if (!px)
          sprintf(szAssemblerFile,"__tmp_%s.S",szInputFile);
        else
        {
          *px=0;
          sprintf(szAssemblerFile,"%s/__tmp_%s.S",szInputFile,px+1);
          *px='/';
        }
        if (!bQuiet)
          fprintf(stdout,"%s: Compiling via AFTERBURNER...\n",argv[0]);
      }
      else
      {
        if (!bQuiet)
          fprintf(stdout,"%s: Compiling...\n",argv[0]);
      }
      iRetVal = DoExecute(argv[0],&argv[0],CompileFile);
      if (iRetVal) 
      {
        if ((bIsICC) && (!bNo2ndTry))
        {
          i=j=1;
          while (argv[i])
          {
            if ((!strcmp(argv[i],"-ip")) || (!memcmp(argv[i],"-ipo",4)))
            {
              if (iOutputPos>i) iOutputPos--;
              if (iCSwitchPos>i) iCSwitchPos--;
              if (iInputPos>i) iInputPos--;
              i++;
            }
            else
            {
              if (!strcmp(argv[i],"-Ob2"))
                argv[i][3]='1';
              argv[j++]=argv[i++];
            }
          }
          argv[j]=NULL;
          if (!bQuiet)
            fprintf(stdout,"%s: Compiling (2nd chance, cause: internal compiler error)...\n",argv[0]);
          iRetVal = DoExecute(argv[0],&argv[0],CompileFile);
          if (iRetVal)
            return iRetVal;
        }
        else
          return iRetVal;
      }

      if (bPreproc)
      {
        if (!bQuiet)
          fprintf(stdout,"%s: Preprocessing...\n",argv[0]);
        iRetVal = DoExecute(argv[0],&argv[0],PreprocessFile);
        if (iRetVal) return iRetVal;
      }

      if (bAssem)
      {
        if (!bQuiet)
          fprintf(stdout,"%s: Assembling...\n",argv[0]);
        if ((bIsICC) && ((bNoCommon) || (bRemoveForceJunk) || (bOptInline) || (bSafeSSE)))
        {
          char szNewFile[256];
          sprintf(szNewFile,"%s.S",szInputFile);
          if (rename(szAssemblerFile,szNewFile))
          {
            if (!bQuiet)
              fprintf(stdout,"%s: Assembling (step 2)...\n",argv[0]);
            unlink(szAssemblerFile);
            iRetVal = DoExecute(argv[0],&argv[0],AssembleFile);
            if (iRetVal) return iRetVal;
          }
        }
        else
        {
          iRetVal = DoExecute(argv[0],&argv[0],AssembleFile);
          if (iRetVal) return iRetVal;
        }
      }
 
      return 0;      
    }
  }
  execvp(argv[0],&argv[0]);
  return 1;
}

