/**
 * file             : iccab.c
 * author           : Ingo A. Kubbilun <ingo@pyrillion.org>
 * license          : GNU GENERAL PUBLIC LICENSE 2 (or later)
 * 
 * description      : This is the icc after burner, part of the
 *                    C compiler delegate ccd.
 *
 *                    See ccd(1) for details.
 *
 * revision history :
 * ------------------
 * 
 * version  date        comments
 * ---------------------------------------------------------------------
 * 0.1      04/01/2004  initial version - Ingo A. Kubbilun
**/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>

#define FILE_SCOPE          200

typedef int (*PCHILDPROC)(char *, char **);
int ICCAfterBurner ( char *argv0, char **argv );
extern int DoExecute ( char *argv0, char **argv, PCHILDPROC pChildProc );

extern char                 szCompiler[64];               // compiler to use, e.g. gcc or icc
extern int                  bIsICC;                       // is it the Intel Compiler for Linux (icc) ?
extern int                  bAssem;
extern int                  bPreproc;
extern int                  bQuiet;
extern int                  bCompileOnly;
extern int                  bOutputGiven;
extern int                  bInputGiven;
extern int                  bInputIsAssembler;
extern int                  bNoCommon;
extern int                  bRemoveForceJunk;
extern int                  bOptInline;
extern int                  bSafeSSE;
extern int                  iCSwitchPos;
extern int                  iOutputPos;
extern int                  iInputPos;
extern int                  bDetectasmBug1;
extern char                 szOutputFile[256];
extern char                 szInputFile[256];
extern char                 szAssemblerFile[256];

typedef struct tagSECSTCK         SECSTCK, *PSECSTCK;

struct tagSECSTCK
{
  PSECSTCK      pPrev;
  char          szSection[128];
};

static PSECSTCK             pSS;
static char                 szCurSection[128];
static unsigned char        szBuffer[4096];
static int                  iLines=0,iLinesAvail=0;
static unsigned char      **ppLines = NULL;

static char                 szArgv0[] = "gcc";
static char                 szArgv1[] = "-O2";
static char                 szArgv2[] = "-c";
static char                 szArgv3[] = "-o";
static char                *ppArgv[8] = { szArgv0,szArgv1,szArgv2,szArgv3,NULL,NULL,NULL,NULL };

static int AssembleFile ( char *argv0, char **argv )
{
  argv[iOutputPos]     = szAssemblerFile;
  argv[iCSwitchPos][1] = 'S';

  execvp(argv0,argv);
  return 1;
}

static unsigned char *IgnWS ( unsigned char *p )
{
  while ((*p) && ((*p)<0x21)) p++;
  return p;
}

static unsigned char *IgnNWS ( unsigned char *p )
{
  while ((*p) && ((*p)>0x20)) p++;
  return p;
}

static int Trim ( unsigned char *pszString, int len )
{
  unsigned char *p;
  int i;

  if (!len) return 0;

  while ((len>0) && (pszString[len-1]<0x21)) len--;
  if (!len) return 0;
  pszString[len]=0;

  p=pszString;
  while ((*p) && ((*p)<0x21)) { p++;len--; }
  if (!len) return 0;

  if (p>pszString)
  {
    i=0;
    while (*p) { pszString[i++] = *(p++); }
    pszString[i]=0;
    return i;
  }  
  return len;
}

static int InsertLine ( int idx, unsigned char *pszNewLine )
{
  unsigned char         **ppNewLines,*pszNew;
  int                     i,iMove,len;

  if (idx>iLines) return 0; // ==iLines: append!

  if ((iLines+1)>iLinesAvail)
  {
    ppNewLines = (unsigned char **)malloc(sizeof(unsigned char*)*(iLinesAvail+1024));
    if (!ppNewLines) return 0;
    memset(ppNewLines,0,sizeof(unsigned char*)*(iLinesAvail+1024));
    memcpy(ppNewLines,ppLines,sizeof(unsigned char *)*iLinesAvail);
    free(ppLines);
    ppLines = ppNewLines;
    iLinesAvail += 1024;
  }

  if ((iMove=iLines-idx)>0)
  {
    for (i=0;i<iMove;i++)
      ppLines[iLines-i] = ppLines[iLines-i-1];
  }

  ppLines[idx] = NULL;
  len    = strlen(pszNewLine)+1;
  pszNew = (unsigned char *)malloc(len);
  if (!pszNew) return 0;
  memcpy(pszNew,pszNewLine,len);
  ppLines[idx] = pszNew;
  iLines++;
  return 1;
}

static int ReplaceLine ( int idx, unsigned char *pszNewLine )
{
  int               len    = strlen(pszNewLine)+1;
  unsigned char    *pszNew = (unsigned char *)malloc(len);

  if (!pszNew) return 0;
  if (idx>=iLines)
  {
    free(pszNew);
    return 0;
  }
  memcpy(pszNew,pszNewLine,len);
  ppLines[idx] = pszNew;
  return 1;
}

static int RemarkLine ( int idx )
{
  if (idx>=iLines) return 0;
  sprintf(szBuffer,"# ***ICCAB REMARK***: %s",ppLines[idx]);
  return ReplaceLine(idx,szBuffer);
}

typedef struct tagCSYM        CSYM, *PCSYM;

struct tagCSYM
{
  PCSYM               pNext;
  unsigned long       uCommonSize;
  unsigned long       uCommonAlign;
  unsigned char       szSymbol[4];
};

static PCSYM              pCRoot=NULL,pCAct=NULL;

static int AddCommonSymbol ( unsigned char *pszSymbol, unsigned long uCommonSize, unsigned long uCommonAlign )
{
  int         len = strlen(pszSymbol);
  PCSYM       pNew = (PCSYM)malloc(sizeof(CSYM)+len);

  if (!pNew) return 0;

  pNew->pNext         = NULL;
  pNew->uCommonSize   = uCommonSize;
  pNew->uCommonAlign  = uCommonAlign;
  memcpy(pNew->szSymbol,pszSymbol,len+1);

  if (!pCRoot)
    pCRoot = pNew;
  else
    pCAct->pNext = pNew;
  pCAct = pNew;

  return 1;
}

static int GCCCompileFile ( char *argv0, char **argv )
{
#ifdef _DEBUG
  int i=0;
  while (argv[i]) fprintf(stdout,"%s ",argv[i++]);
  fprintf(stdout,"\n");
#endif
  execvp(argv0,argv);
  return 1;
}

int ICCAfterBurner ( char *argv0, char **argv )
{
  int                     iRetVal,iSize,i,j,idx,len,iCSymbols=0,iForceJunkFuncs=0,i1,i2;
  FILE                   *hFile     = NULL;
  FILE                   *hDiag     = NULL;
  unsigned char          *pbySource = NULL;
  unsigned char          *p,*p2,*p3,*p4;
  unsigned char           szCommonSymbol[256],szSrcReg[256],szCachedReg[256];
  unsigned long           uCommonAlign,uCommonSize;
  unsigned char           c;
  unsigned char           szFunction[256];
  unsigned char           szJunkFunc[256];
  PSECSTCK                pNewS;
  static unsigned char    szCurrentLine[1024];
  char                    szTemp[128];
  int                     isInitExit,isText,iType;
  unsigned long           uValue,uAlign;

  memset(szCachedReg,0,sizeof(szCachedReg));

  // 1.) Compile the C source to assembler source

  iRetVal = DoExecute(argv[0],&argv[0],AssembleFile);
  if (iRetVal) 
  {
    unlink(szAssemblerFile);
    return iRetVal;
  }
    
  // 2.) Read the assembler source entirely

  hFile = fopen(szAssemblerFile,"r");
  if (!hFile)
  {
    AccError:
    if (pbySource) free(pbySource);
    if (hFile) fclose(hFile);
    fprintf(stderr,"CCC(iccab) FATAL ERROR: cannot access temporary assembler source. ABORTING.\n");
    fflush(stderr);
    unlink(szAssemblerFile);
    return 1;
  }
  fseek(hFile,0L,SEEK_END);
  iSize = ftell(hFile);
  fseek(hFile,0L,SEEK_SET);
  if (!iSize)
  {
    fclose(hFile);
    fprintf(stderr,"CCC(iccab) FATAL ERROR: temporary assembler source is empty. ABORTING.\n");
    fflush(stderr);
    unlink(szAssemblerFile);
    return 1;
  }
  pbySource = (unsigned char *)malloc(iSize+16);
  if (!pbySource)
  {
    fclose(hFile);
    fprintf(stderr,"CCC(iccab) FATAL ERROR: malloc error. ABORTING.\n");
    fflush(stderr);
    unlink(szAssemblerFile);
    return 1;
  }
  if (fread(pbySource,1,iSize,hFile)!=iSize)
    goto AccError;
  fclose(hFile);
  hFile=NULL;

  // 3.) Count number of lines

  if (pbySource[iSize-1]!='\n')
    pbySource[iSize++]='\n';

  for (i=0;i<iSize;i++)
  {
    if (pbySource[i]=='\0')
    {
      free(pbySource);
      fprintf(stderr,"CCC(iccab) FATAL ERROR: temporary assembler source contains zero(s). ABORTING.\n");
      fflush(stderr);
      unlink(szAssemblerFile);
      return 1;
    }
    else
    if (pbySource[i]=='\n')
    {
      pbySource[i]='\0';
      iLines++;
    }
  }

  iLinesAvail = (iLines<<1);
  ppLines = (unsigned char **)malloc(sizeof(unsigned char*)*iLinesAvail);
  if (!ppLines)
  {
//    MallocError:
    free(pbySource);
    if (ppLines) free(ppLines);
    fprintf(stderr,"CCC(iccab) FATAL ERROR: malloc error. ABORTING.\n");
    fflush(stderr);
    unlink(szAssemblerFile);
    return 1;
  }
  memset(ppLines,0,sizeof(unsigned char*)*iLinesAvail);

  idx=0;
  ppLines[idx++] = pbySource;
  for (i=0;i<(iSize-1);i++)
  {
    if (!pbySource[i])
      ppLines[idx++] = pbySource+i+1;
  }
  ppLines[iLines]=NULL;
  
  if (strcmp(ppLines[iLines-1],"# End"))
  {
    free(pbySource);
    free(ppLines);
    fprintf(stderr,"CCC(iccab) FATAL ERROR: end directive not found. ABORTING.\n");
    fflush(stderr);
    unlink(szAssemblerFile);
    return 1;
  }

  // 4.) Remove __F_O_R_C_E__xxx functions

  if (bRemoveForceJunk)
  {
    i=0;
    while (i<iLines)
    {
      if (!memcmp(ppLines[i],"# -- Begin  __F_O_R_C_E__",25))
      {
        p  = IgnNWS(ppLines[i]+25);
        c  = *p;
        *p = 0;
        strcpy(szFunction,ppLines[i]+25);
        *p = c;
        sprintf(szJunkFunc,"# -- End  __F_O_R_C_E__%s",szFunction);
        j = i+1;
        while (j<iLines)
        {
          if (!strcmp(ppLines[j],szJunkFunc))
            break;
          else
            j++;
        }
        if (j<iLines)
        {
          while (i<=j)
          {
            p=IgnWS(ppLines[i]);
            if ((!memcmp(p,".text",5)) || (!memcmp(p,".data",5)) || (!memcmp(p,".bss",4)) || (!memcmp(p,".sec",4)))
              i++;
            else
              RemarkLine(i++);
          }
          iForceJunkFuncs++;
        }
        else
          i++;
      }
      else
        i++;
    }
  }

  // 5.) Eliminate common variables

  if (bNoCommon)
  {
    for (i=0;i<iLines;i++)
    {
      p = IgnWS(ppLines[i]);
      if ((!memcmp(p,".comm",5)) && (p[5]<0x21))
      {
        p=IgnWS(p+5);
        p2 = strchr(p,',');
        if (p2)
        {
          len = ((unsigned long)p2)-((unsigned long)p);
          memcpy(szCommonSymbol,p,len);
          szCommonSymbol[len]=0;
          if (Trim(szCommonSymbol,len))
          {
            p = IgnWS(p2+1);
            uCommonSize = strtoul(p,(char**)&p,0);
            if ((uCommonSize) && (p))
            {
              uCommonAlign=0;
              p = IgnWS(p);
              if ((*p)==',')
                uCommonAlign = strtoul(IgnWS(p+1),(char**)&p,0);

              RemarkLine(i);
              AddCommonSymbol(szCommonSymbol,uCommonSize,uCommonAlign);
            }
          }
        }
      }
    }

    // convert common symbols

    if (pCRoot)
    {
      InsertLine(iLines-1,"\t.bss");
      while (pCRoot)
      {
        iCSymbols++;
        sprintf(szBuffer,"\t.globl %s",pCRoot->szSymbol);
        InsertLine(iLines-1,szBuffer);
        if (pCRoot->uCommonAlign)
        {
          sprintf(szBuffer,"\t.align %lu",pCRoot->uCommonAlign);
          InsertLine(iLines-1,szBuffer);
        }
        sprintf(szBuffer,"\t.type %s, @object",pCRoot->szSymbol);
        InsertLine(iLines-1,szBuffer);
        sprintf(szBuffer,"\t.size %s, %lu",pCRoot->szSymbol,pCRoot->uCommonSize);
        InsertLine(iLines-1,szBuffer);
        sprintf(szBuffer,"%s:",pCRoot->szSymbol);
        InsertLine(iLines-1,szBuffer);
        sprintf(szBuffer,"\t.zero %lu",pCRoot->uCommonSize);
        InsertLine(iLines-1,szBuffer);

        pCAct = pCRoot;
        pCRoot = pCRoot->pNext;
        free(pCAct);
      }
    }
  }

  // Disable PGO calls in sections ".init.text" and ".exit.text"

  strcpy(szCurSection,".text");
  isInitExit = 0;
  isText     = 1;
  pSS = NULL;
  for (i=0;i<iLines;i++)
  {
    strcpy(szCurrentLine,ppLines[i]);
    p=strchr(szCurrentLine,'#');
    if (p) *p=0;
    if ((NULL!=(p=strstr(szCurrentLine,".text"))) && ((p==szCurrentLine) || (p[-1]<0x21)))
    {
      pNewS = (PSECSTCK)malloc(sizeof(SECSTCK));
      if (!pNewS)
      {
        SSErr:
        fprintf(stderr,"CCC(iccab) FATAL ERROR: malloc(). ABORTING.\n");
        fflush(stderr);
        unlink(szAssemblerFile);
        return 1;
      }
      memset(pNewS,0,sizeof(SECSTCK));
      strcpy(pNewS->szSection,szCurSection);
      pNewS->pPrev = pSS;
      pSS          = pNewS;
      strcpy(szCurSection,".text");
      isText       = 1;
      isInitExit   = 0;
    }
    else
    if (strstr(szCurrentLine,".data"))
    {
      pNewS = (PSECSTCK)malloc(sizeof(SECSTCK));
      if (!pNewS) goto SSErr;
      memset(pNewS,0,sizeof(SECSTCK));
      strcpy(pNewS->szSection,szCurSection);
      pNewS->pPrev = pSS;
      pSS          = pNewS;
      strcpy(szCurSection,".data");
      isText       = 0;
      isInitExit   = 0;
    }
    else
    if (strstr(szCurrentLine,".bss"))
    {
      pNewS = (PSECSTCK)malloc(sizeof(SECSTCK));
      if (!pNewS) goto SSErr;
      memset(pNewS,0,sizeof(SECSTCK));
      strcpy(pNewS->szSection,szCurSection);
      pNewS->pPrev = pSS;
      pSS          = pNewS;
      strcpy(szCurSection,".bss");
      isText       = 0;
      isInitExit   = 0;
    }
    else
    if (NULL!=(p=strstr(szCurrentLine,".section")))
    {
      p=IgnWS(p+sizeof(".section")-1);
      if ((*p) && (p2=IgnNWS(p)) && (p2>p))
      {
        DoSection:
        p3=strchr(p,',');
        if ((p3) && (p3<p2)) p2=p3;
        pNewS = (PSECSTCK)malloc(sizeof(SECSTCK));
        if (!pNewS) goto SSErr;
        memset(pNewS,0,sizeof(SECSTCK));
        strcpy(pNewS->szSection,szCurSection);
        pNewS->pPrev = pSS;
        pSS          = pNewS;
        c=*p2;
        *p2=0;
        strcpy(szCurSection,p);
        *p2=c;
        DoEval:
        isText = 0;
        if ((!strcmp(szCurSection,".init.text")) || (!strcmp(szCurSection,".exit.text")))
          isInitExit = 1;
        else
        {
          isInitExit = 0;
          if (!strcmp(szCurSection,".text"))
            isText = 1;
        }
      }
    }
    else
    if (NULL!=(p=strstr(szCurrentLine,".pushsection")))
    {
      p=IgnWS(p+sizeof(".pushsection")-1);
      if ((*p) && (p2=IgnNWS(p)) && (p2>p))
        goto DoSection;
    }
    else
    if (NULL!=(p=strstr(szCurrentLine,".popsection")))
    {
      if (pSS)
      {
        strcpy(szCurSection,pSS->szSection);
        pNewS = pSS;
        pSS   = pSS->pPrev;
        free(pNewS);
        goto DoEval;
      }
      else
      {
        strcpy(szCurSection,".text");
        isText       = 1;
        isInitExit   = 0;
      }
    }
    else
    if (NULL!=(p=strstr(szCurrentLine,".previous")))
    {
      if (pSS)
      {
        strcpy(szTemp,szCurSection);
        strcpy(szCurSection,pSS->szSection);
        strcpy(pSS->szSection,szTemp);
        goto DoEval;
      }
    }

    if (isInitExit)
    {
      p=strstr(szCurrentLine,"call");
      if (p)
      {
        p2=strstr(p+4,"_PGOPTI_Prof_Begin");
        if (p2)
        {
          RemarkLine(i);
        }
      }
    }
    else
    if (isText)
    {
      p=strstr(szCurrentLine,".align");
      if (p)
      {
        p=IgnWS(p+6);
        uAlign=strtoul(p,(char**)&p,0);
        if ((uAlign) && (p))
        {
          p=IgnWS(p);
          if ((*p)==',')
          {
            p=IgnWS(p+1);
            uValue=strtoul(p,(char**)&p,0);
            if (uValue==0x90)
            {
              switch(uAlign)
              {
                case 2:
                  uValue=1;
                  DoP2:
                  sprintf(szCurrentLine,"\t.p2align\t%lu\t\t# ICCAB align replacement",uValue);
                  ReplaceLine(i,szCurrentLine);
                  break;
                case 4:
                  uValue=2;
                  goto DoP2;
                case 8:
                  uValue=3;
                  goto DoP2;
                case 16:
                  uValue=4;
                  goto DoP2;
                case 32:
                  uValue=5;
                  goto DoP2;
                case 64:
                  uValue=6;
                  goto DoP2;
                case 128:
                  uValue=7;
                  goto DoP2;
                case 256:
                  uValue=8;
                  goto DoP2;
                default:
                  break;
              }
            }
          }
        }
      }
    }
  }

  // handle asm bug1

  if (bDetectasmBug1)
  {
    i=0;
    while (i<iLines)
    {
      if (strstr(ppLines[i],"# Begin ASM"))
      {
        i++;
        while (i<iLines)
        {
          if (strstr(ppLines[i],"# End ASM"))
            break;
          if (((p=strstr(ppLines[i],"btsl"))!=NULL) ||
              ((p=strstr(ppLines[i],"btcl"))!=NULL) ||
              ((p=strstr(ppLines[i],"btrl"))!=NULL) ||
              ((p=strstr(ppLines[i],"btl"))!=NULL))
          {
            if (strstr(p,"(%esp"))
            {
              fprintf(stderr,"Suspicious bit test construct found in %s\n",szInputFile);
              i1 = i-3;
              i2 = i+3;
              if (i1<0) i1=0;
              if (i2>(iLines-1)) i2=(iLines-1);
              hDiag = fopen("diag.log","at");
              if (hDiag)
              {
                fprintf(hDiag,"\nfile %s:\n",szInputFile);
                for (j=i1;j<=i2;j++)
                  fprintf(hDiag,"%s\n",ppLines[j]);
                fclose(hDiag);
              }
            }
            else
            if ((p=strstr(p,",0"))!=NULL)
            {
              fprintf(stderr,"WRONG bit test construct found in %s\n",szInputFile);
              i1 = i-3;
              i2 = i+3;
              if (i1<0) i1=0;
              if (i2>(iLines-1)) i2=(iLines-1);
              hDiag = fopen("diagerror.log","at");
              if (hDiag)
              {
                fprintf(hDiag,"\nfile %s:\n",szInputFile);
                for (j=i1;j<=i2;j++)
                  fprintf(hDiag,"%s\n",ppLines[j]);
                fclose(hDiag);
              }

              iType = 1;  // backward
              i1 = i-2;
              i2 = i-(FILE_SCOPE+2);
              if (i2<0) i2=0;
              while (i1>i2)
              {
                if ((strstr(ppLines[i1],"movl")) && (strstr(ppLines[i1+1],"testl")) && (!strstr(ppLines[i1],"__pgo_segment")))
                  break;
                i1--;
              }
              if (i1==i2)
              {
                iType = 0;  // forward
                i1 = i+1;
                i2 = i+(FILE_SCOPE+1);
                if (i2>(iLines-2)) i2=iLines-2;
                while (i1<i2)
                {
                  if ((strstr(ppLines[i1],"movl")) && (strstr(ppLines[i1+1],"testl")) && (!strstr(ppLines[i1],"__pgo_segment")))
                    break;
                  i1++;
                }
                if (i1==i2)
                {
                  iType = 2; // cached
                  if (!szCachedReg[0]) goto UnableToCorrect;
                  p4 = szCachedReg;
                  goto DoPatchReg;
                }
              }
              if (i1==i2)
              {
                UnableToCorrect:
                fprintf(stderr,"    => CANNOT CORRECT.\n");
              }
              else
              {
                fprintf(stderr,"%s\n%s\n",ppLines[i1],ppLines[i1+1]);
                p3 = strstr(ppLines[i1],", %");
                p4 = strstr(ppLines[i1+1],", %");
                if ((!p3) || (!p4)) goto UnableToCorrect;
                if (memcmp(p3,p4,6)) goto UnableToCorrect;
                (*p3)=0;
                strcpy(szSrcReg,ppLines[i1]);
                (*p3)=',';
                p3 = szSrcReg;
                p4 = p3+strlen(szSrcReg)-1;
                while ((p4>=p3) && (((unsigned char)(*p4))>0x20)) p4--;
                p4++;
                DoPatchReg:
                (*p)=0;
                strcpy(szCurrentLine,ppLines[i]);
                (*p)=',';
                strcat(szCurrentLine,",");
                strcat(szCurrentLine,p4);
                if (p4!=szCachedReg) strcpy(szCachedReg,p4);
                switch(iType)
                {
                  case 0 :strcat(szCurrentLine,"\t# ICCAB: error corrected (FORWARD)");break;
                  case 1 :strcat(szCurrentLine,"\t# ICCAB: error corrected (BACKWARD)");break;
                  default:strcat(szCurrentLine,"\t# ICCAB: error corrected (CACHED)");break;
                }
                ReplaceLine(i,szCurrentLine);
                fprintf(stderr,"    => CORRECTED.\n");
              }
            }
          }
          i++;
        }
        i++;
      }
      else
        i++;
    }    
  }

  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  hFile = fopen(szAssemblerFile,"w");
  if (!hFile)
  {
    WBError:
    if (hFile) fclose(hFile);
    fprintf(stderr,"CCC(iccab) FATAL ERROR: cannot write back temporary assembler source. ABORTING.\n");
    fflush(stderr);
    unlink(szAssemblerFile);
    return 1;
  }

  for (i=0;i<iLines;i++)
  {
    if (ppLines[i]==NULL)
      goto WBError;
    fprintf(hFile,"%s\n",ppLines[i]);
  }

  fclose(hFile);

  ppArgv[4] = argv[iOutputPos];
  ppArgv[5] = szAssemblerFile;

  iRetVal = DoExecute(ppArgv[0],ppArgv,GCCCompileFile);
  if (iRetVal) return iRetVal;

  if (!bAssem)
    unlink(szAssemblerFile);
  return 0;
}






