/**
 * file             : pgod.c
 * author           : Ingo A. Kubbilun <ingo@pyrillion.org>
 * license          : GNU GENERAL PUBLIC LICENSE 2 (or later)
 * 
 * description      : This is the user mode part of the PGO kernel
 *                    module, the PGO daemon pgod.
 *
 *                    See pgod(8) for details.
 *
 * revision history :
 * ------------------
 * 
 * version  date        comments
 * ---------------------------------------------------------------------
 * 0.1      04/01/2004  initial version - Ingo A. Kubbilun
**/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mount.h>

#define VAR_LOCK          "/var/lock/subsys/pgod"
#define VAR_RUN           "/var/run/pgod.pid"
#define PGO_DIR           "/var/PGO"
#define DEFAULT_INTERVAL  30
#define ACTION_SHUTDOWN   1L
#define ACTION_STOP       2L
#define ACTION_CONTINUE   4L

static int                g_iCmd          =  0;
static int                g_iInterval     =  DEFAULT_INTERVAL;
static int                g_iVerbose      =  0;
static char               g_szProfDir[256];
static pid_t              g_PID           =  0;
static int                g_lockfd        = -1;
static unsigned long      g_uActionMask   =  0L;
static int                g_EPM           =  0;
static pid_t              g_WaitForPID    =  0;
static char               szArg0[]        = "profmerge";
static char               szArg1[]        = "-prof_dir";
static char              *pArgv[4]        = { szArg0,szArg1,g_szProfDir,NULL };
static char               g_szDynFile[256];

static int ExecProfMerge ( void );
static int SendDaemonSignal ( int signum );
static int GetValue ( char *pszValue );
static int PutValue ( char *pszValue, int iValue );
static void SignalHandler ( int sig, siginfo_t *pSI, void *pContext );
static void LogMessage ( int loglevel, const char *pcszFormat, ... );
static void DoDaemonWork ( void );
static void ClearAllBuffers ( void );

static void DoShutdown ( void )
{
  DoDaemonWork();

  umount(PGO_DIR);
  rmdir(PGO_DIR);

  if (g_lockfd!=-1)
  {
    lockf(g_lockfd,F_ULOCK,0);
    close(g_lockfd);
    unlink(VAR_LOCK);
  }
  unlink(VAR_RUN);
  LogMessage(LOG_INFO,"pgod terminated normally.\n");
  exit(0);
}

static int SendDaemonSignal ( int signum )
{
  FILE               *hFile = fopen(VAR_RUN,"rt");
  pid_t               pid;

  if (!hFile) return -1;

  if ((1!=fscanf(hFile,"%i",&pid)) || (!pid))
  {
    fclose(hFile);
    return -1;
  }
  fclose(hFile);

  return kill(pid,signum);
}

static int GetValue ( char *pszValue )
{
  char        szFile[256];
  FILE       *hFile;
  int         iValue;

  sprintf(szFile,"%s/%s",PGO_DIR,pszValue);
  hFile = fopen(szFile,"rt");
  if (!hFile) return -1;
  if (1!=fscanf(hFile,"%i",&iValue))
  {
    fclose(hFile);
    return -1;
  }
  fclose(hFile);
  return iValue;
}

static int PutValue ( char *pszValue, int iValue )
{
  char        szFile[256],szString[256];
  FILE       *hFile;

  sprintf(szFile,"%s/%s",PGO_DIR,pszValue);
  hFile = fopen(szFile,"wt");
  if (!hFile) return -1;
  sprintf(szString,"%i\n",iValue);
  if (strlen(szString)!=fprintf(hFile,"%s",szString))
  {
    fclose(hFile);
    return -1;
  }
  fclose(hFile);
  return 0;
}

static void ClearAllBuffers ( void )
{
  int i, iBuffers = GetValue("buffers");

  for (i=0;i<iBuffers;i++) PutValue("clear",1);

  PutValue("reset",1);
}

typedef struct tagMFLIST      MFLIST, *PMFLIST;

struct tagMFLIST
{
  PMFLIST         pNext;
  char            szFile[256];
};

static PMFLIST    g_pRoot = NULL, g_pCurr = NULL;

static void AddToList ( void )
{
  PMFLIST       pNew = (PMFLIST)malloc(sizeof(MFLIST));

  if (!pNew) return;

  strncpy(pNew->szFile,g_szDynFile,sizeof(pNew->szFile));
  pNew->pNext = NULL;

  if (!g_pRoot)
    g_pRoot = pNew;
  else
    g_pCurr->pNext = pNew;

  g_pCurr = pNew;
}

static int ExecProfMerge ( void )
{
  int           pipefd[2],status=127;
  FILE         *fp;
  char         *p, *p2;
  pid_t         waitpid;
  char          szOutput[256];

  memset(g_szDynFile,0,sizeof(g_szDynFile));

	if (pipe(pipefd)<0) 
    return -1;

	if ((g_WaitForPID=fork())<0)
  {
    g_WaitForPID = 0;
    close(pipefd[0]);
    close(pipefd[1]);
    return -1;
  }

  if (!g_WaitForPID)  // child
  {
	  close(pipefd[0]);
		if (pipefd[1] != STDERR_FILENO) 
    {
		  dup2(pipefd[1], STDERR_FILENO);
		  close(pipefd[1]);
		}
    execvp(pArgv[0],pArgv);
    exit(127);
  }
  else                // parent
  {
		close(pipefd[1]);
		if ((fp=fdopen(pipefd[0],"r"))==NULL)
    {
      g_WaitForPID=0;
      close(pipefd[0]);
      return -1;
    }
    while (fgets(szOutput,sizeof(szOutput),fp))
    {
      p=strstr(szOutput,"profmerge:");
      if (p)
      {
        p2 = strstr(p+sizeof("profmerge:")-1,"merging dynamic file:");
        if (p2)
        {
          p2 += sizeof("merging dynamic file:")-1;
          while ((*p2) && (((unsigned char)(*p2))<0x21)) p2++;
          p=p2;
          while ((*p) && (((unsigned char)(*p))>0x20)) p++;
          *p=0;
          strncpy(g_szDynFile,p2,sizeof(g_szDynFile));
          AddToList();
          memset(g_szDynFile,0,sizeof(g_szDynFile));
        }
      }
    }
    fclose(fp);

    while ((waitpid=wait4(g_WaitForPID,&status,0,NULL))<0)
    {
  		if (errno != EINTR)
			  break;
    }

    if ((waitpid!=g_WaitForPID) || (WIFSIGNALED(status)) || (!WIFEXITED(status)))
    {
      g_WaitForPID=0;
      return -1;
    }   

    g_WaitForPID=0;

    return WEXITSTATUS(status);
  }

  return -1;
}
 
static void DumpBuffer ( int iSize )
{
  void               *pBuffer = malloc(iSize);
  long                l;
  char                szFilename[512];
  FILE               *hFile;
  int                 status;
  PMFLIST             pTemp;

  if (!pBuffer)
  {
    LogMessage(LOG_ERR,"malloc() error (%i bytes required). PGO buffer discarded.\n",iSize);
    return;
  }

  hFile = fopen(PGO_DIR "/dyndata","rb");
  if (!hFile)
  {
    LogMessage(LOG_ERR,"cannot open PGO data. PGO buffer discarded.\n");
    free(pBuffer);
    return;
  }

  if (fread(pBuffer,1,iSize,hFile)!=iSize)
  {
    LogMessage(LOG_ERR,"cannot read PGO data. PGO buffer discarded.\n");
    fclose(hFile);
    free(pBuffer);
    return;
  }
  fclose(hFile);

  l = (long)time(NULL);
  FilenameLoop:
  sprintf(szFilename,"%s%08lx_%05d.dyn",g_szProfDir,l,g_PID);
  hFile = fopen(szFilename,"r");
  if (hFile)
  {
    fclose(hFile);
    l++;
    goto FilenameLoop;
  }

  LogMessage(LOG_INFO,"about to dump PGO buffer (%i bytes) to %s.\n",iSize,szFilename);

  hFile = fopen(szFilename,"wb");
  if (!hFile)
  {
    LogMessage(LOG_ERR,"cannot create PGO file. PGO buffer discarded.\n");
    free(pBuffer);
    return;
  }

  if (fwrite(pBuffer,1,iSize,hFile)!=iSize)
  {
    LogMessage(LOG_ERR,"cannot dump PGO data to file. PGO buffer discarded.\n");
    fclose(hFile);
    free(pBuffer);
    return;
  }

  fclose(hFile);
  free(pBuffer);

  if (g_EPM)
  {
    status = ExecProfMerge();
    if (status)
    {
      LogMessage(LOG_ERR,"cannot execute profmerge tool. PGO buffer remains on disk.\n");
    }
    else
    {
      while (g_pRoot)
      {
        sprintf(szFilename,"%s%s",g_szProfDir,g_pRoot->szFile);
        LogMessage(LOG_INFO,"profmerge merged %s. PGO buffer deleted.\n",szFilename);
        unlink(szFilename);
        pTemp   = g_pRoot;
        g_pRoot = g_pRoot->pNext;
        free(pTemp);
      }
      g_pCurr = NULL;
    }
  }
}

static void DoDaemonWork ( void )
{
  int             i,iBuffers,iSize;

  // 1.) Process old buffers

  ChkBuffers:
  iBuffers = GetValue("buffers");
  if (iBuffers<0)
  {
    LogMessage(LOG_ERR,"cannot get number of active buffers.\n");
    return;
  }
  LogMessage(LOG_INFO,"number of buffers not yet processed is %i.\n",iBuffers);

  if (iBuffers>0)
  {
    for (i=0;i<iBuffers;i++)
    {
      iSize = GetValue("dynsize");
      if (iSize<0)
        LogMessage(LOG_ERR,"cannot get PGO buffer size.\n");
      else
      {
        if (iSize>8)
          DumpBuffer(iSize);
        else
          LogMessage(LOG_INFO,"buffer dump canceled (cause: buffer does not contain data).\n");
      }
      PutValue("clear",1);
    }  
    goto ChkBuffers;
  }

  if (-1==PutValue("dump",1))
    LogMessage(LOG_ERR,"cannot trigger PGO data dump.\n");
  else
  {
    iSize = GetValue("dynsize");
    if (iSize<0)
      LogMessage(LOG_ERR,"cannot get PGO buffer size.\n");
    else
    {
      if (iSize>8)
        DumpBuffer(iSize);
      else
        LogMessage(LOG_INFO,"buffer dump canceled (cause: buffer does not contain data).\n");
    }
    PutValue("clear",1);
  }
}
 
static void SignalHandler ( int sig, siginfo_t *pSI, void *pContext )
{
  pid_t               pid;
  int                 status;

  switch(sig)
  {
    case SIGHUP:      // re-read configuration
      break;

    case SIGALRM:     // timer expired
      break;

    case SIGUSR1:     // user1 (stop PGO kernel module)
      g_uActionMask |= ACTION_STOP;
      break;

    case SIGUSR2:     // user2 (continue PGO kernel module)
      g_uActionMask |= ACTION_CONTINUE;
      break;

    case SIGSTOP:     // stop
      break;

    case SIGCONT:     // continue
      break;

    case SIGCHLD:     // child terminated
      if (g_WaitForPID<=0)
      {
        pid=wait4(0,&status,WNOHANG,NULL);
        g_WaitForPID=0;
      }
      break;

    case SIGTERM:     // terminate daemon
      g_uActionMask |= ACTION_SHUTDOWN;
      break;

    default:          // everything else: ignore
      break;
  }
}

static void LogMessage ( int loglevel, const char *pcszFormat, ... )
{
  va_list       val;
  static char   szBuffer[4096];

  if ((loglevel==LOG_INFO) && (!g_iVerbose)) return;

  va_start(val,pcszFormat);
  vsprintf(szBuffer,pcszFormat,val);
  va_end(val);

  openlog("pgod",LOG_PID,LOG_DAEMON);
	syslog(loglevel,"%s",szBuffer);
  closelog();
}

int main ( int argc, char *argv[] )
{
  char                   *p = getenv("PROF_DIR");
  int                     i,l,iRetVal=1,stpcnt=-1,iResetOnInit=0,iStop=0;
  pid_t                   pid;
  struct sigaction        sa;
  FILE                   *hFile;
  time_t                  lasttime,acttime;

  if (p)
    strcpy(g_szProfDir,p);
  else
    strcpy(g_szProfDir,"/profdata");

  l=strlen(g_szProfDir);
  if ((l) && (g_szProfDir[l-1]!='/')) strcat(g_szProfDir,"/");

  p = getenv("PROF_INTERVAL");
  if (p)
  {
    g_iInterval = atoi(p);
    if (g_iInterval<=0) g_iInterval = DEFAULT_INTERVAL;
  }

  for (i=1;i<argc;i++)
  {
    if (!strcmp(argv[i],"--start"))
      g_iCmd = 0;
    else
    if (!strcmp(argv[i],"--kill"))
      g_iCmd = 1;
    else
    if (!strcmp(argv[i],"--reset"))
      g_iCmd = 2;
    else
    if (!strcmp(argv[i],"--reset-on-init"))
      iResetOnInit = 1;
    else
    if (!strcmp(argv[i],"--exec-profmerge"))
      g_EPM = 1;
    else
    if (!strcmp(argv[i],"--stop"))
      g_iCmd = 3;
    else
    if (!strcmp(argv[i],"--cont"))
      g_iCmd = 4;
    else
    if (!memcmp(argv[i],"--prof_dir=",11))
    {
      strcpy(g_szProfDir,argv[i]+11);
      l=strlen(g_szProfDir);
      if ((l) && (g_szProfDir[l-1]!='/')) strcat(g_szProfDir,"/");
    }
    else
    if (!memcmp(argv[i],"--prof_interval=",16))
    {
      g_iInterval = atoi(argv[i]+16);
      if (g_iInterval<=0) g_iInterval = DEFAULT_INTERVAL;
    }
    else
    if (!strcmp(argv[i],"--verbose"))
      g_iVerbose = 1;
    else
    {
      fprintf(stderr,"unknown switch '%s'.\n",argv[i]);
      return 1;
    }
  }

  if (!g_iCmd)
  {
    pid = fork();
    if (pid<0) 
    {
      fprintf(stderr,"ERROR: cannot fork.\n");
      return 1;
    }

    if (pid>0) return 0;

    setsid();

    for (i=getdtablesize();i>=0;--i) close(i);
    i = open("/dev/null",O_RDWR);
  	dup(i);                      
  	dup(i);                      

    umask(027);

    chdir("/");

    g_lockfd = open(VAR_LOCK,O_RDWR|O_CREAT,0640);

    LogMessage(LOG_INFO,"pgod startup...");

    if (g_lockfd<0) 
    {
      LogMessage(LOG_ERR," => pgod startup failed.\n");
      exit(1);
    }

    if (lockf(g_lockfd,F_TLOCK,0)<0) 
    {
      LogMessage(LOG_INFO," => pgod already running.\n");
      exit(0);
    }

    g_PID = getpid();

    hFile = fopen(VAR_RUN,"wt");
  
    if (hFile)
    {
      fprintf(hFile,"%i\n",g_PID);
      fclose(hFile);
    }

    memset(&sa,0,sizeof(sa));
    sa.sa_flags = SA_SIGINFO|SA_RESTART;
    sa.sa_sigaction = SignalHandler;
    sigemptyset(&sa.sa_mask);
    sigaction(SIGHUP,&sa,NULL);
    sigaction(SIGTERM,&sa,NULL);
    sigaction(SIGALRM,&sa,NULL);
    sigaction(SIGUSR1,&sa,NULL);
    sigaction(SIGUSR2,&sa,NULL);
    sigaction(SIGCONT,&sa,NULL);
    sigaction(SIGSTOP,&sa,NULL);
    sa.sa_flags |= SA_NOCLDSTOP;
    sigaction(SIGCHLD,&sa,NULL);
    sa.sa_flags = 0;
    sa.sa_handler = SIG_IGN;
    sigaction(SIGINT,&sa,NULL);
    sigaction(SIGQUIT,&sa,NULL);
    sigaction(SIGPIPE,&sa,NULL);
    sigaction(SIGTSTP,&sa,NULL);
    sigaction(SIGTTIN,&sa,NULL);
    sigaction(SIGTTOU,&sa,NULL);

    mkdir(PGO_DIR,0700);
    if (mount("pgofs",PGO_DIR,"pgofs",0xC0ED0000L|MS_NOEXEC|MS_SYNCHRONOUS|MS_NOATIME|MS_NODIRATIME,NULL))
    {
      LogMessage(LOG_ERR,"FATAL: unable to mount PGO filesystem. Killing daemon.\n");
      exit(1);
    }

    LogMessage(LOG_INFO," => operational mode (prof dir=%s, interval=%i sec(s).\n",g_szProfDir,g_iInterval);
    if (iResetOnInit)
      ClearAllBuffers();

    lasttime=time(NULL);
    DoDaemonWork();

    DaemonLoop:
    if (g_uActionMask & ACTION_SHUTDOWN)
      DoShutdown();
    if (g_uActionMask & ACTION_STOP)
    {
      g_uActionMask &= ~ACTION_STOP;
      iStop++;
    }
    if (g_uActionMask & ACTION_CONTINUE)
    {
      g_uActionMask &= ~ACTION_CONTINUE;
      iStop--;
      if (iStop<0) iStop=0;
    }
    sleep(1);
    if (iStop>0)
      goto DaemonLoop;

    acttime=time(NULL);
    if (((int)difftime(acttime,lasttime))>=g_iInterval)
    {
      lasttime=acttime;
      DoDaemonWork();
    }
    goto DaemonLoop;
  }
  else
  {
    switch(g_iCmd)
    {
      case 1: // kill
        iRetVal = SendDaemonSignal(SIGTERM);
        fprintf(stdout,"PGO control: command 'kill' %ssent to PGO daemon.\n",iRetVal ? "NOT " : "");
        break;

      case 2: // reset
        iRetVal = PutValue("reset",1);
        fprintf(stdout,"PGO control: command 'reset' %ssent to kernel module.\n",iRetVal ? "NOT " : "");
        break;

      case 3: // stop
        iRetVal = SendDaemonSignal(SIGUSR1);
        fprintf(stdout,"PGO control: command 'stop' %ssent to PGO daemon.\n",iRetVal ? "NOT " : "");
        iRetVal = PutValue("stop",1);
        if (!iRetVal)
          stpcnt = GetValue("stop");
        fprintf(stdout,"PGO control: command 'stop' %ssent to kernel module.\n",iRetVal ? "NOT " : "");
        if (stpcnt!=-1)
          fprintf(stdout,"             current stop count is %i.\n",stpcnt);
        else
          fprintf(stdout,"             current stop count is UNKNOWN.\n");
        break;

      case 4: // cont
        iRetVal = SendDaemonSignal(SIGUSR2);
        fprintf(stdout,"PGO control: command 'continue' %ssent to PGO daemon.\n",iRetVal ? "NOT " : "");
        iRetVal = PutValue("stop",0);
        if (!iRetVal)
          stpcnt = GetValue("stop");
        fprintf(stdout,"PGO control: command 'continue' %ssent to kernel module.\n",iRetVal ? "NOT " : "");
        if (stpcnt!=-1)
          fprintf(stdout,"             current stop count is %i%s.\n",stpcnt,!stpcnt ? " (running)" : " (still stopped)");
        else
          fprintf(stdout,"             current stop count is UNKNOWN.\n");
        break;

      default:
        break;
    }
  }

  return iRetVal ? 1 : 0;
}
